from __future__ import annotations

import numpy as np
import sasktran2 as sk2
import xarray as xr
from scipy.linalg import block_diag

from skretrieval.retrieval.prior import BasePrior

from . import StateVectorElement


class StateVectorElementConstituent(
    StateVectorElement, sk2.constituent.base.Constituent
):
    def __init__(
        self,
        constituent: sk2.constituent.base.Constituent,
        constituent_name: str,
        property_names: list[str],
        min_value=None,
        max_value=None,
        prior: dict[BasePrior] | None = None,
        log_space=False,
        enabled=True,
    ):
        """
        A state vector element that is a sasktran2.constituent

        Parameters
        ----------
        constituent : sk2.constituent.base.Constituent
            The sasktran2 constituent
        constituent_name : str
            A name for the constituent
        property_names : list[str]
            Property names of the constituent that will be retrieved
        min_value : dict, optional
            Minimum values for the property names as a dictionary, by default {}
        max_value : dict, optional
            maximumum values for the property names as a dictionary, by default {}
        prior : dict, optional
            Prior objects for each property name, by default {}
        log_space : bool, optional
            If true then the state elements will be rescaled to logarithmic space, by default False
        """
        if prior is None:
            prior = {}
        if max_value is None:
            max_value = {}
        if min_value is None:
            min_value = {}
        self._log_space = log_space
        self._constituent = constituent
        self._property_names = property_names
        self._constituent_name = constituent_name
        self._min_value = min_value
        self._max_value = max_value

        self._prior = prior

        start = 0
        for property_name in self._property_names:
            if property_name in self._prior:
                n = len(getattr(self._constituent, property_name))
                self._prior[property_name].init(self, slice(start, start + n))
                start += n
            else:
                self._prior[property_name] = BasePrior()

        super().__init__(enabled)

    def state(self) -> np.array:
        data = []

        for property_name in self._property_names:
            data.append(getattr(self._constituent, property_name))

        if self._log_space:
            return np.log(np.hstack(data))
        return np.hstack(data)

    def lower_bound(self) -> np.array:
        data = []
        for property_name in self._property_names:
            x = getattr(self._constituent, property_name)

            bound = self._min_value.get(property_name, -np.inf)

            data.append(np.ones(len(x)) * bound)

        if self._log_space:
            return np.log(np.hstack(data))
        return np.hstack(data)

    def upper_bound(self) -> np.array:
        data = []
        for property_name in self._property_names:
            x = getattr(self._constituent, property_name)

            bound = self._max_value.get(property_name, np.inf)

            data.append(np.ones(len(x)) * bound)

        if self._log_space:
            return np.log(np.hstack(data))
        return np.hstack(data)

    def inverse_apriori_covariance(self) -> np.ndarray:
        prior_mats = []

        for property_name in self._property_names:
            inv_S_a = self._prior[property_name].inverse_covariance

            if self._log_space:
                prior_mats.append(inv_S_a)
            else:
                prior_mats.append(
                    inv_S_a
                    / np.outer(
                        self._prior[property_name].state,
                        self._prior[property_name].state,
                    )
                )

        return block_diag(*prior_mats)

    def apriori_state(self) -> np.array:
        return np.concatenate(
            [self._prior[property].state for property in self._property_names]
        )

    def name(self) -> str:
        return self._constituent_name

    def propagate_wf(self, radiance: xr.Dataset) -> xr.Dataset:
        if "extinction_per_m" in self._property_names:
            radiance = radiance.rename(
                {
                    f"wf_{self._constituent_name}_extinction": f"wf_{self._constituent_name}_extinction_per_m"
                }
            )
        wfs = []
        for property_name in self._property_names:
            wfs.append(
                radiance[f"wf_{self._constituent_name}_{property_name}"].rename(
                    {
                        radiance[f"wf_{self._constituent_name}_{property_name}"].dims[
                            0
                        ]: "x"
                    }
                )
            )

            if self._log_space:
                x = getattr(self._constituent, property_name)
                wfs[-1].values *= x[:, np.newaxis, np.newaxis, np.newaxis]

        return xr.concat(wfs, dim="x")

    def update_state(self, x: np.array):
        start = 0
        for property_name in self._property_names:
            current = getattr(self._constituent, property_name)
            property_length = len(current)
            if self._log_space:
                sv = np.exp(x[start : start + property_length])
                sv[np.isnan(sv)] = self._max_value[property_name]
            else:
                sv = x[start : start + property_length]
            if property_name in self._min_value:
                sv[sv < self._min_value[property_name]] = self._min_value[property_name]
            if property_name in self._max_value:
                sv[sv > self._max_value[property_name]] = self._max_value[property_name]

            self._constituent.__setattr__(property_name, sv)

            start += property_length

    def modify_input_radiance(self, radiance: xr.Dataset):
        return radiance

    def add_to_atmosphere(self, atmo: sk2.Atmosphere):
        return self._constituent.add_to_atmosphere(atmo)

    def register_derivative(self, atmo: sk2.Atmosphere, name: str):
        return self._constituent.register_derivative(atmo, name)

    def adjust_constituent_attributes(self, **kwargs):
        for key, value in kwargs.items():
            if isinstance(value, dict):
                for k, v in value.items():
                    if k.lower() == "scale":
                        setattr(
                            self._constituent, key, getattr(self._constituent, key) * v
                        )
                    if k.lower() == "set":
                        setattr(self._constituent, key, v)
            else:
                setattr(self._constituent, key, getattr(self._constituent, key) * value)

    def describe(self, **kwargs) -> xr.Dataset | None:
        ds = xr.Dataset()

        if (
            type(self._constituent)
            is sk2.constituent.brdf.lambertiansurface.LambertianSurface
        ):
            albedo = getattr(self._constituent, self._property_names[0])

            ds[self._constituent_name] = xr.DataArray(
                albedo,
                dims=[self._constituent._interp_var],
                coords={self._constituent._interp_var: self._constituent._x},
            )
            ds[self._constituent_name + "_1sigma_error"] = xr.DataArray(
                np.sqrt(np.diag(kwargs["covariance"])),
                dims=[self._constituent._interp_var],
                coords={self._constituent._interp_var: self._constituent._x},
            )

        else:
            start = 0
            for property_name in self._property_names:
                end = start + len(getattr(self._constituent, property_name))

                ds[self._constituent_name + "_" + property_name] = xr.DataArray(
                    getattr(self._constituent, property_name), dims=["altitude"]
                )

                ds[
                    self._constituent_name + "_" + property_name + "_prior"
                ] = xr.DataArray(self._prior[property_name].state, dims=["altitude"])

                if "covariance" in kwargs:
                    if self._log_space:
                        ds[
                            self._constituent_name
                            + "_"
                            + property_name
                            + "_1sigma_error"
                        ] = xr.DataArray(
                            np.sqrt(np.diag(kwargs["covariance"])[start:end])
                            * getattr(self._constituent, property_name),
                            dims=["altitude"],
                        )
                    else:
                        ds[
                            self._constituent_name
                            + "_"
                            + property_name
                            + "_1sigma_error"
                        ] = xr.DataArray(
                            np.sqrt(np.diag(kwargs["covariance"])[start:end]),
                            dims=["altitude"],
                        )

                if "averaging_kernel" in kwargs:
                    ds[
                        self._constituent_name
                        + "_"
                        + property_name
                        + "_averaging_kernel"
                    ] = xr.DataArray(
                        kwargs["averaging_kernel"][start:end, start:end],
                        dims=["altitude", "altitude_2"],
                    )

                start = end

        return ds



from __future__ import annotations

import numpy as np
from scipy import sparse


def estimate_error(
    K: np.ndarray,
    Sy: np.ndarray,
    inv_Sy: np.ndarray,
    inv_Sa: np.ndarray,
    left_side_eqn: np.ndarray | None = None,
) -> dict:
    """
    Estimates the error and averaging kernel for the retrieval process

    Parameters
    ----------
    K : np.ndarray
        Jacobian matrix
    Sy : np.ndarray
        Instrument error covariance matrix
    inv_Sy : np.ndarray
        Invers of the instrument error covariance matrix
    inv_Sa : np.ndarray
        Inverse of the a priori error covariance matrix
    left_side_eqn : np.ndarray | None, optional
        Left side of the retrieval equation, (K.T @ inv_Sy @ K + inv_Sa), by default None.
        If set to None it is calculated by this function.

    Returns
    -------
    dict
        Dictionary containing the following keys:
        - gain_matrix: Gain matrix
        - averaging_kernel: Averaging kernel
        - error_covariance_from_noise: Error covariance from noise
        - solution_covariance: Solution covariance
    """
    output_dict = {}

    if left_side_eqn is not None:
        A_without_lm = left_side_eqn
    else:
        A_without_lm = K.T @ inv_Sy @ K + inv_Sa

    # Calculate the solution covariance and averaging kernels
    try:
        if sparse.issparse(A_without_lm):
            S = np.linalg.inv(A_without_lm.toarray())
        else:
            S = np.linalg.inv(A_without_lm)
    except np.linalg.LinAlgError:
        if sparse.issparse(A_without_lm):
            S = np.linalg.pinv(A_without_lm.toarray())
        else:
            S = np.linalg.pinv(A_without_lm)

    G = S @ K.T @ inv_Sy
    A = G @ K
    meas_error_covar = G @ (Sy.dot(G.T))

    output_dict["gain_matrix"] = G
    output_dict["averaging_kernel"] = A

    output_dict["error_covariance_from_noise"] = meas_error_covar
    output_dict["solution_covariance"] = S

    return output_dict



from __future__ import annotations


def plot_state(results: dict, state: str, **kwargs):
    import matplotlib.pyplot as plt

    r = results["state"]
    plt.subplot(1, 2, 1)
    plt.plot(r[f"{state}"], r["altitude"])
    plt.plot(r[f"{state}_prior"], r["altitude"])
    plt.legend(["Retrieved", "Prior"])

    plt.ylabel("Altitude [m]")
    plt.xlabel(f"{state}")

    plt.subplot(1, 2, 2)
    plt.plot(r[f"{state}_averaging_kernel"], r["altitude"])
    plt.xlabel("Averaging Kernel")
    plt.ylabel("Altitude [m]")

    plt.tight_layout()

    if kwargs.get("show", False):
        plt.show()



from __future__ import annotations

import numpy as np
import sasktran2 as sk2
import xarray as xr

from skretrieval.retrieval.statevector import StateVector
from skretrieval.retrieval.statevector.constituent import StateVectorElementConstituent


class AltitudeNativeStateVector(StateVector):
    def __init__(self, altitude_grid: np.array, **kwargs):
        """
        Class representing the state vector where state vector elements are natively
        specified on altitude levels

        Parameters
        ----------
        altitude_grid : np.array
            Native altitude grid that state vector elements should be specified on. Often
            matched to the SASKTRAN2 model grid
        """
        self._sv_eles = kwargs
        self._altitude_grid = altitude_grid

        elements = []

        for _, val in self._sv_eles.items():
            elements.append(val)

        super().__init__(elements)

    def add_to_atmosphere(self, atmo: sk2.Atmosphere):
        """
        Add`s the state vector elements to the atmosphere object

        Parameters
        ----------
        atmo : sk2.Atmosphere
        """
        for key, val in self._sv_eles.items():
            if isinstance(val, StateVectorElementConstituent):
                atmo[key] = val

    @property
    def sv(self):
        """
        The state vector elements

        Returns
        -------
        dict[StateVectorElement]
        """
        return self._sv_eles

    @property
    def altitude_grid(self):
        """
        The altitude grid that the state vector elements are defined on in [m]

        Returns
        -------
        np.array
        """
        return self._altitude_grid

    def post_process_sk2_radiances(self, radiance: xr.Dataset):
        """
        Called after the SASKTRAN2 radiance calculation to modify the radiance if the state vector elements
        require it.

        Parameters
        ----------
        radiance : xr.Dataset
        """
        wfs = []
        for val in self.state_elements:
            if val.enabled:
                wfs.append(val.propagate_wf(radiance))

        for val in self.state_elements:
            if val.enabled:
                radiance = val.modify_input_radiance(radiance)

        wfs = xr.concat(wfs, dim="x")

        wf_keys = []
        for key in list(radiance):
            if key.startswith("wf"):
                wf_keys.append(key)

        radiance = radiance.drop_vars(wf_keys)
        radiance["wf"] = wfs

        return radiance

    def describe(self, rodgers_output: dict) -> xr.Dataset:
        """
        Returns a human readable dataset containing the state vector elements

        Parameters
        ----------
        rodgers_output : dict

        Returns
        -------
        xr.Dataset
        """
        result = super().describe(rodgers_output)

        result.coords["altitude"] = self._altitude_grid

        return result
